#Fitting a Poisson HMM to given data



nmax <- 50 #upper limit later used for approximate computations


#Log-likelihood of Poisson HMM:
llphmm <- function(par,data){
#par is vector with vectors (lambda,Acol0,...,Acold), where diagonal elements of A are omitted (expected to be large)
	T <- length(data)
	minval <- min(data)
	maxval <- max(data)

	nostates <- sqrt(length(par)) #d+1

	lambdas <- par[1:nostates]

	#Transition matrix of states:
	A <- array(0, c(nostates,nostates))
	for(s in c(1:nostates)){
		acol <- par[(nostates+(s-1)*(nostates-1)+1):(nostates+s*(nostates-1))] #elements of column s without diagonal element
		acol <- append(acol,1-sum(acol),(s-1)) #insert diagonal element after position s-1
		A[,s] <- acol
	}

	#Vector of marginal probabilities from Markov property:
	pmarg <- eigen(A, symmetric = FALSE)$vectors[,1]
	pmarg <- Re(pmarg/sum(pmarg))

	#For computational reasons, preparing required matrices for alpha computation:
	matPA <- array(0, c((maxval-minval+1), nostates,nostates))
	for(x in c(minval:maxval)){
		matPA[x-minval+1,,] <- diag(dpois(x, lambdas)) %*% A
	}

	#Computing full likelihood function via alphas:
	value <- diag(dpois(data[1], lambdas)) %*% pmarg

	for(t in c(2:T)) {
		value <- matPA[data[t]-minval+1,,] %*% value
		#If numerical problems, then use scheme based on  (wt/wt1, phit), see CatHMMfit.r.
	}
	-log(sum(value))
}



#If working with optimization with inequality constraints (ui %*% parvec >= ci), 
#where parvec is vector with vectors (lambda,Acol0,...,Acold), where diagonal elements of A are omitted (expected to be large),
#the following matrix and vector are required:
constrmat <- function(nostates){
	nopar <- nostates + nostates*(nostates-1)
	mat <- array(0, c(nopar+nostates,nopar))
	
	#All parameters >=0:
	mat[1:nopar,1:nopar] <- diag(1, nopar)
	
	#Column sums <=1:
	for(j in c(1:nostates)){
		pos <- nostates + (j-1)*(nostates-1)
		mat[nopar+j, (pos+1):(pos+nostates-1)] <- (-1)
	}
	
	mat
}

constrvec <- function(nostates){
	nopar <- nostates + nostates*(nostates-1)
	vec <- rep(0, nopar+nostates)
	
	#Column sums <=1:
	vec[(nopar+1):(nopar+nostates)] <- rep(-1, nostates)
	
	vec
}




data <- scan(file.choose()) #Downloads.txt
Tlen <- length(data)
Tlen #267

maxval <- max(data)
maxval #14

plot(data, type="b", pch=19, cex=0.5, xlab = "t", ylab = expression("Download counts  x"[t]), ylim=c(0,maxval), cex.axis=0.85)

acf(data)

pacf(data)

rho1 <- acf(data, plot=FALSE)[[1]][2]
rho1 #0.2447806


#Absolute frequencies
absfreq <- tabulate(data+1) #+1 to include 0
plot(0:maxval, absfreq/Tlen, type="h", xlab = "k", ylab = expression(paste("estimated P(X"[t],"=k)")), lwd=4, ylim=c(0,0.3))

#Observations' mean
barX <- mean(data)
barX
# 2.400749

#Observations' variance
sX <- var(data)
sX
#7.534285

(Tlen-1)/Tlen*sX
#7.506067

#Testing for equidispersion:
ID <- (Tlen-1)/Tlen*sX/barX
ID #3.126552

meanID <- 1-1/Tlen*(1+rho1)/(1-rho1)
meanID #0.9938268

sdID <- sqrt(2/Tlen*(1+rho1^2)/(1-rho1^2))
sdID #0.09189935

#Upper critical value:
meanID+qnorm(0.95)*sdID
#1.144988

#Upper-sided P-value:
1-pnorm(ID, mean=meanID, sd=sdID)
#0




#Fitting 2-state Poisson HMM
nostates <- 2

#First the (slow) way with inequality constraints (ui %*% parvec >= ci):
suppressWarnings(constrOptim(c(c(1,5),c(0.2,0.4)), llphmm, NULL, ui=constrmat(nostates), ci=constrvec(nostates), data=data))
# $par
# [1] 0.9953333 5.2670082 0.1965659 0.3947783
# $value
# [1] 548.9348




#Now with box constraints allowing to approximate standard errors, 
#which certainly requires an appropriate choice of the box lengths:
estml <- suppressWarnings(optim(c(c(1,5),c(0.2,0.4)), llphmm, method="L-BFGS-B", lower=c(0.5,1.5,0.05,0.2), upper=c(3,6,0.4999,0.4999), control=list(ndeps=rep(1e-4,4)), data=data, hessian=TRUE))

lambdasestml <- estml$par[1:nostates]
lambdasestml
#0.9952829 5.2667765

Aestml <- array(0, c(nostates,nostates))
for(s in c(1:nostates)){ #insert diagonal elements as above
	acol <- estml$par[(nostates+(s-1)*(nostates-1)+1):(nostates+s*(nostates-1))]
	acol <- append(acol,1-sum(acol),(s-1))
	Aestml[,s] <- acol
}
Aestml
#          [,1]      [,2]
#[1,] 0.8033342 0.3948333
#[2,] 0.1966658 0.6051667

pmargestml <- eigen(Aestml, symmetric = FALSE)$vectors[,1]
pmargestml <- Re(pmargestml/sum(pmargestml))
pmargestml
#0.667513 0.332487


ofiest <- estml$hessian #inverse covariance
neglmax <- estml$value
estcov <- solve(ofiest)

#Estimated standard errors:
c(sqrt(diag(estcov)))
#0.12732495 0.40897245 0.03991422 0.08017038


#AIC and BIC:
AIC <- 2*neglmax+2* nostates^2
BIC <- 2*neglmax+log(Tlen)*nostates^2
c(neglmax, AIC, BIC)
#548.9348 1105.8696 1120.2186






#Properties of fitted model

#Mean
muestml <- c(lambdasestml %*% pmargestml)
muestml #2.415499

#Variance
sig2estml <- c(((lambdasestml+lambdasestml^2) %*% pmargestml) - muestml^2)
sig2estml #6.464929 #disp 2.676

#PMF:
pmfestml <- rep(0, (maxval+1))
for(x in c(0:maxval)){
	pmfestml[x+1] <- sum(diag(dpois(x,lambdasestml)) %*% pmargestml)
}
#pmfestml
plot(0:maxval, absfreq/Tlen, type="h", xlab = "k", ylab = expression(paste("estimated P(X"[t],"=k)")), lwd=4, ylim=c(0,0.3))
points((0:maxval)+.25,pmfestml, type="h", lwd=4, col=gray(0.5))

#ACF
maxlag <- 10
rhosestml <- rep(0,maxlag)

Ak <- diag(1,nostates)
for(k in c(1:maxlag)){
	Ak <- Aestml %*% Ak
	rhosestml[k] <- c(lambdasestml %*% Ak %*% (lambdasestml * pmargestml))
	rhosestml[k] <- (rhosestml[k]-muestml^2)/sig2estml
}
acf(data, plot=FALSE)[[1]][2:(maxlag+1)]
#[1]0.24478064 0.08243551 0.08624810 0.06431224 0.08089815 0.08281390 0.04191897 0.06868241 0.08247258 0.03109923
rhosestml
#[1] 2.558723e-01 1.045241e-01 4.269819e-02 1.744225e-02 7.125176e-03 2.910641e-03 1.189000e-03 4.857076e-04 1.984120e-04 8.105151e-05






#Pearson residuals

#Compute all alphas first
alphas <- array(0, c(Tlen,nostates))

alphas[1,] <- c(diag(dpois(data[1], lambdasestml)) %*% pmargestml)
for(t in c(2:Tlen)) {
	alphas[t,] <- c(diag(dpois(data[t], lambdasestml)) %*% Aestml %*% alphas[t-1,])
}
#log(sum(alphas[Tlen,]))
#-548.9348

#1-step-ahead distributions
cmeans <- rep(0,Tlen)
cvars <- rep(1,Tlen)

for(t in c(2:Tlen)){
	cdist <- rep(0,nmax+1)
	for(x in c(0:nmax)){
		cdist[x+1] <- sum(diag(dpois(x, lambdasestml)) %*% Aestml %*% alphas[t-1,])/sum(alphas[t-1,])
	}
	cmeans[t] <- c(cdist %*% c(0:nmax))
	cvars[t] <- c(cdist %*% (c(0:nmax)^2))-cmeans[t]^2
}

#Pearson residuals
res <- (data-cmeans)/sqrt(cvars)
res <- res[-1]
acf(res) #OK
plot(res)
mean(res)
#-0.01686642
var(res)
#1.112754
min(res)
#-1.270593
max(res)
#4.611153





#PIT:
nobins <- 10
PIT <- array(0, c(2,nobins+1))

for(t in c(2:Tlen)){
	ft <- sum(diag(dpois(0, lambdasestml)) %*% Aestml %*% alphas[t-1,])/sum(alphas[t-1,])
	ft1 <- 0
	if(data[t]>0){
		for(x in c(1:data[t])){
			ft1 <- ft
			ft <- ft + sum(diag(dpois(x, lambdasestml)) %*% Aestml %*% alphas[t-1,])/sum(alphas[t-1,])
		}
	}

	for(j in c(1:nobins)){
		u <- j/nobins
		pitval <- 0
		if(ft1<u){
			if(ft<u){
				pitval <- 1
			}else{
				pitval <- (u-ft1)/(ft-ft1)
			}
		}
	PIT[1,j+1] <- PIT[1,j+1] + pitval
	}
}
PIT[1,] <- PIT[1,]/(Tlen-1)
PIT[2,2:(nobins+1)] <- PIT[1,2:(nobins+1)]-PIT[1,1:nobins]
PIT
     # [,1]      [,2]      [,3]      [,4]       [,5]      [,6]       [,7]
# [1,]    0 0.1085397 0.2155863 0.3184496 0.41533237 0.5238727 0.61687470
# [2,]    0 0.1085397 0.1070466 0.1028634 0.09688273 0.1085403 0.09300201
           # [,8]       [,9]      [,10]      [,11]
# [1,] 0.71484748 0.81340521 0.90600418 1.00000000
# [2,] 0.09797278 0.09855773 0.09259897 0.09399582

PIT.freq= as.vector(rep(((1:nobins)-0.5)/nobins, PIT[2,2:(nobins+1)]*1000))
PIT.hist <- hist(PIT.freq, plot=FALSE, breaks=nobins)
PIT.hist$counts <- PIT.hist$counts/1000
plot(PIT.hist, freq=TRUE, main="PIT histogram", ylab="PIT histogram", xlab="u", col="gray")










#Fitting 3-state Poisson HMM
nostates <- 3

#First the (slow) way with inequality constraints (ui %*% parvec >= ci):
suppressWarnings(constrOptim(c(c(1,3,8),c(0.25,0.05,0.3,0.1,0.15,0.5)), llphmm, NULL, ui=constrmat(nostates), ci=constrvec(nostates), data=data))
# $par
# [1] 0.63091753 2.97898853 8.22348869 0.24474503 0.04226289 0.28299579 0.11782862
# [8] 0.15776798 0.50128232
# $value
# [1] 536.1962




#Now with box constraints allowing to approximate standard errors, 
#which certainly requires an appropriate choice of the box lengths:
estml <- suppressWarnings(optim(c(c(1,3,8),c(0.25,0.05,0.3,0.1,0.15,0.5)), llphmm, method="L-BFGS-B", lower=c(0.5,1.5,3,rep(1e-2,6)), upper=c(2,5,10,0.4,0.4,0.4,0.4,0.25,0.7), control=list(ndeps=rep(1e-4,9)), data=data, hessian=TRUE))

lambdasestml <- estml$par[1:nostates]
lambdasestml
#0.6309787 2.9794816 8.2304387

Aestml <- array(0, c(nostates,nostates))
for(s in c(1:nostates)){ #insert diagonal elements as above
	acol <- estml$par[(nostates+(s-1)*(nostates-1)+1):(nostates+s*(nostates-1))]
	acol <- append(acol,1-sum(acol),(s-1))
	Aestml[,s] <- acol
}
Aestml
           # [,1]      [,2]      [,3]
# [1,] 0.71299803 0.2829667 0.1575307
# [2,] 0.24478183 0.5992815 0.5019420
# [3,] 0.04222014 0.1177518 0.3405273

pmargestml <- eigen(Aestml, symmetric = FALSE)$vectors[,1]
pmargestml <- Re(pmargestml/sum(pmargestml))
pmargestml
#0.4732393 0.4212478 0.1055130


ofiest <- estml$hessian #inverse covariance
neglmax <- estml$value
estcov <- solve(ofiest)

#Estimated standard errors:
c(sqrt(diag(estcov)))
#[1] 0.15133884 0.42560466 0.91739869 0.08548399 0.03622894 0.06877198 0.06289605 0.13103683 0.16934139


#AIC and BIC:
AIC <- 2*neglmax+2* nostates^2
BIC <- 2*neglmax+log(Tlen)*nostates^2
c(neglmax, AIC, BIC)
#536.1963 1090.3925 1122.6778




#Properties of fitted model

#Mean
muestml <- c(lambdasestml %*% pmargestml)
muestml #2.422122

#Variance
sig2estml <- c(((lambdasestml+lambdasestml^2) %*% pmargestml) - muestml^2)
sig2estml #7.630868 #disp 3.150

#PMF:
pmfestml <- rep(0, (maxval+1))
for(x in c(0:maxval)){
	pmfestml[x+1] <- sum(diag(dpois(x,lambdasestml)) %*% pmargestml)
}
#pmfestml
plot(0:maxval, absfreq/Tlen, type="h", xlab = "k", ylab = expression(paste("estimated P(X"[t],"=k)")), lwd=4, ylim=c(0,0.3))
points((0:maxval)+.25,pmfestml, type="h", lwd=4, col=gray(0.5))


#ACF
maxlag <- 10
rhosestml <- rep(0,maxlag)

Ak <- diag(1,nostates)
for(k in c(1:maxlag)){
	Ak <- Aestml %*% Ak
	rhosestml[k] <- c(lambdasestml %*% Ak %*% (lambdasestml * pmargestml))
	rhosestml[k] <- (rhosestml[k]-muestml^2)/sig2estml
}
acf(data, plot=FALSE)[[1]][2:(maxlag+1)]
#[1] 0.24478064 0.08243551 0.08624810 0.06431224 0.08089815 0.08281390 0.04191897 0.06868241 0.08247258 0.03109923
rhosestml
#[1] 0.2738463527 0.1198435574 0.0545946182 0.0252941186 0.0117992356 0.0055190802 0.0025843118 0.0012106160 0.0005672050 0.0002657676





#Pearson residuals

#Compute all alphas first
alphas <- array(0, c(Tlen,nostates))

alphas[1,] <- c(diag(dpois(data[1], lambdasestml)) %*% pmargestml)
for(t in c(2:Tlen)) {
	alphas[t,] <- c(diag(dpois(data[t], lambdasestml)) %*% Aestml %*% alphas[t-1,])
}
#log(sum(alphas[Tlen,]))
#-536.1963

#1-step-ahead distributions
cmeans <- rep(0,Tlen)
cvars <- rep(1,Tlen)

for(t in c(2:Tlen)){
	cdist <- rep(0,nmax+1)
	for(x in c(0:nmax)){
		cdist[x+1] <- sum(diag(dpois(x, lambdasestml)) %*% Aestml %*% alphas[t-1,])/sum(alphas[t-1,])
	}
	cmeans[t] <- c(cdist %*% c(0:nmax))
	cvars[t] <- c(cdist %*% (c(0:nmax)^2))-cmeans[t]^2
}

#Pearson residuals
res <- (data-cmeans)/sqrt(cvars)
res <- res[-1]
acf(res) #OK
plot(res)
mean(res)
#-0.01441121
var(res)
#0.9689774
min(res)
#-1.235495
max(res)
#4.476367





#PIT:
nobins <- 10
PIT <- array(0, c(2,nobins+1))

for(t in c(2:Tlen)){
	ft <- sum(diag(dpois(0, lambdasestml)) %*% Aestml %*% alphas[t-1,])/sum(alphas[t-1,])
	ft1 <- 0
	if(data[t]>0){
		for(x in c(1:data[t])){
			ft1 <- ft
			ft <- ft + sum(diag(dpois(x, lambdasestml)) %*% Aestml %*% alphas[t-1,])/sum(alphas[t-1,])
		}
	}

	for(j in c(1:nobins)){
		u <- j/nobins
		pitval <- 0
		if(ft1<u){
			if(ft<u){
				pitval <- 1
			}else{
				pitval <- (u-ft1)/(ft-ft1)
			}
		}
	PIT[1,j+1] <- PIT[1,j+1] + pitval
	}
}
PIT[1,] <- PIT[1,]/(Tlen-1)
PIT[2,2:(nobins+1)] <- PIT[1,2:(nobins+1)]-PIT[1,1:nobins]
PIT
     # [,1]      [,2]      [,3]      [,4]       [,5]      [,6]      [,7]
# [1,]    0 0.1011236 0.2018946 0.3011130 0.40029848 0.5080297 0.6132296
# [2,]    0 0.1011236 0.1007709 0.0992184 0.09918551 0.1077312 0.1051999
          # [,8]       [,9]     [,10]      [,11]
# [1,] 0.7091504 0.79937836 0.9054951 1.00000000
# [2,] 0.0959208 0.09022792 0.1061167 0.09450494


PIT.freq= as.vector(rep(((1:nobins)-0.5)/nobins, PIT[2,2:(nobins+1)]*1000))
PIT.hist <- hist(PIT.freq, plot=FALSE, breaks=nobins)
PIT.hist$counts <- PIT.hist$counts/1000
plot(PIT.hist, freq=TRUE, main="PIT histogram", ylab="PIT histogram", xlab="u", col="gray")
plot(PIT.hist, freq=TRUE, main="PIT histogram", ylab="PIT histogram", xlab="u", col="gray", ylim=c(0,0.25))




#Decoding the hidden states

#Compute all betas first
betas <- array(0, c(Tlen,nostates))

betas[Tlen,] <- rep(1,nostates)
for(t in c((Tlen-1):1)) {
	betas[t,] <- c(betas[t+1,] %*% diag(dpois(data[t+1], lambdasestml)) %*% Aestml)
}
#log(c(betas[1,] %*% alphas[1,]))
#-536.1963


#Local decoding:
ll <- sum(alphas[Tlen,])

localstates <- rep(0,Tlen)
for(t in c(1:Tlen)){
	localstates[t] <- which.max(alphas[t,]*betas[t,]/ll)
}
#localstates

plot(data, type="n", xlab = "t", ylab = expression("Download counts  x"[t]), ylim=c(0,maxval), cex.axis=0.85)
points(lambdasestml[localstates], type="p", col=gray(0.5), cex=1, pch=15)
points(data, type="b", pch=19, cex=0.3)



#Global decoding:

#Compute all ms first
ms <- array(0, c(Tlen,nostates))

ms[1,] <- c(pmargestml * dpois(data[1], lambdasestml))
for(t in c(1:(Tlen-1))) {
	ms[t+1,] <- dpois(data[t+1], lambdasestml)
	for(s in c(1:nostates)){
		ms[t+1,s] <- ms[t+1,s] * max(ms[t,]*Aestml[s,])
	}
}

globalstates <- rep(0,Tlen)
globalstates[Tlen] <- which.max(ms[Tlen,])
for(t in c((Tlen-1):1)) {
	globalstates[t] <- which.max(ms[t,]*Aestml[globalstates[t+1],])
}
#globalstates

#Number states being decoded differently:
sum(rep(1,Tlen)[globalstates-localstates!=0])
#3
c(1:Tlen)[globalstates-localstates!=0]
#[1] 142 144 145
localstates[c(1:Tlen)[globalstates-localstates!=0]]
#[1] 2 2 2
globalstates[c(1:Tlen)[globalstates-localstates!=0]]
#[1] 1 1 1

plot(data, type="n", xlab = "t", ylab = expression("Download counts  x"[t]), ylim=c(0,maxval), cex.axis=0.85)
points(lambdasestml[globalstates], type="p", col=gray(0.5), cex=1, pch=15)
points(data, type="b", pch=19, cex=0.3)

